/////////////////////////////////////////////////////////////////////
// tdlmain.cpp
//
// IEEE 1451.4 Template Description Language (TDL) syntax checker.
//
// 27 January 2003
// Updated 10 September 2003
//
// This program uses code generated by flex and bison to parse
// TDL templates as defined by the IEEE 1451.4 standard.  The files
// used as input to flex and bison are tdl.l and tdl.y.  Those files
// contain formal token and grammar definitions.
//
// The main objective of this program is to provide a syntax checker
// for templates.  However, it could also be used as a basis for a
// TEDS parser, or as a portion of a T-Block implementation.
//
// The basic design is simply to parse through the template a token
// at a time.  At appropriate places in the parse, usually at the end
// of a template command, a semantic action routine is invoked.  These
// routines mostly just output the line number, the command type, and
// any parameters associated with the command.  An indication of when
// to read bits from the TEDS and some other minor activity suggestions
// are occasionaly provided.
//
// Note that this program only does a rudimentary check of the syntax 
// of the quoted format field in property commands.
//
// Property command names are verified against the initial property
// list defined in the standard.  Any unrecognized properties are
// listed at the end of the message file.
//
// The minimum and maximum number of bits for each template is output
// at the end of the message file.  However, because of the complexity
// of calculating options regarding selectcase statements, string types,
// free form commands, align commands, etc., there is no guarentee that
// these calculations are correct.  Indeed, some of the commands do not
// allow an exact calculation to be made.
//
// Other C++ files needed are: 
//
//	tdlparser.h - the main header file declaring global variables
//	tdlparser.cpp - the semantic action routines that implement
//	processing of the template commands.
//
// Inputs:
//
//	The template file name and associated template file.  The default 
//	filename is "Default.tdl".
//  The file property_list.txt. This is a text file containing all
//  the properties, on property name per line including the %.
//
// Outputs: 
//
//	messages.dat - A file containing descriptions of each statement
//	in the template file.
//						
//	tokens.dat - A file containing the parsed tokens.  This is most
//	useful for debugging since it provides a specific indication of
//	what tokens were parsed.
// 
// Running the program.
// --------------------
//
// The design is really meant to be run from a command prompt with
// the input template file as the one and only argument.  For example:
//
//   >tdl ext_funct.tdl
//
// Alternatively, if it is run without a command line argument, or if
// it is run by double clicking on the executable in Windows, then
// you will be prompted for the input template file.  Hitting return
// at the prompt will cause the program to read from the file default.tdl.
//
// Coding Environment
// ------------------
//
// This code was originally written on Windows 2000 using Microsoft
// Visual C++ Version 6. A reasonable attempt was made to make the
// code generic so that it should be platform independent, but
// there are no guarentees to this effect.
//
// WARRANTY DISCLAIMER. 
// --------------------
//
// There are no guarantees of any kind associated with use of this 
// code and the user uses it at their own risk.
//
/////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <afx.h> 
#include <afxcoll.h> 

extern int yyparse();	// standard yacc/bison routine
extern FILE *yyin;		// standard yacc/bison input file

FILE *tokens;			// output file containing tokens
FILE *messages;			// output file with all messages
int statement_count = 1;


/////////////////////////////////////////////////////////////////////
// Main function
/////////////////////////////////////////////////////////////////////
void main (int argc, char **argv)

{
  char filename[100]; 

  /////////////////////////////////////////
  // Open the messages file
  /////////////////////////////////////////

  strcpy (filename, "message.dat");
  if ( (messages=fopen(filename, "w")) == NULL) {
    printf ("ERROR: TDL Parser Main: Can not open file %s\n", filename);
    fprintf (messages, "ERROR: TDL Parser Main: Can not open file %s\n", filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  };

  printf ("TDL Parser program.\n");
  fprintf (messages, "TDL Parser program.\n");

  /////////////////////////////////////////
  // Open the token file
  /////////////////////////////////////////

  strcpy (filename, "tokens.dat");
  if ( (tokens=fopen(filename, "w")) == NULL) {
    printf ("ERROR: TDL Parser Main: Can not open file %s\n", filename);
    fprintf (messages, "ERROR: TDL Parser Main: Can not open file %s\n", filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  };

  fprintf (tokens, "TDL Parser program.\n");

  /////////////////////////////////////////
  // Open the input file
  /////////////////////////////////////////

  if (argc == 1) {
    printf ("Template file name [default.tdl]?");
    gets (filename);
    if (strcmp (filename, "") == 0) {
      strcpy (filename, "default.tdl");
    }
  }else{
    strcpy (filename, argv[1]);
  };

  printf ("Reading from file '%s'.\n", filename);
  fprintf (messages, "Reading from file '%s'.\n", filename);
  fprintf (tokens, "Reading from file '%s'.\n", filename);

  yyin=fopen(filename, "r");
  if (yyin == NULL) {
    perror(filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  }

  //////////////////////////////////////
  // parse the file
  //////////////////////////////////////

  yyparse();
  fclose(yyin);
 
  fclose (tokens);
  fprintf (messages, "\nDone\n");
  fclose (messages);
  printf ("Done\n");
  
  printf ("Hit any key to exit.");
  getchar();

} // end main