/////////////////////////////////////////////////////////////////////
// tdlparser.cpp
//
// IEEE 1451.4 TDL parser semantic action routines.
//
// 27 January 2003
// Updated 10 September 2003 for second ballot of draft.
//
// These routines work in conjuction with tdlmain.cpp, tdlparser.h,
// tdl.l, and tdl.y.  These last two files are inputs to flex and
// bison which generate lexyy.cpp, tdl_tab.h, and tdl_tab.cpp.
// The routines contained in this file are identified in tdl.y and
// are then generated into tdl_tab.cpp.
//
// The intent of these routines is to illustrate the
// basic flow of the parser as it reads through a template.
// Comments describing this flow are output to the message.dat file.
// These comments include indicating how many bits are read
// and what other actions should have taken place at that point
// during the parse.
//
// The basic structure of these routines is for there to be an
// action routine for each TDL command along with some support
// routines for these commands.  The routines mostly identify the
// particular command and output the associated parameters that were
// read from the template.  There are also start and end
// routines for the template file itself and for each template.
//
// Someone intending to use this program as a basis for a larger
// program would spend most of their efforts in modifying the
// routines in this file.  All of the information needed for each
// command is available within the routine that processes that command.
// One particular function that would have to be performed would be
// to translate numeric values into numeric variables since everything
// is handled as strings in these routines.
//
// One global function included in these routines is to verify property
// names (this is not done in the grammar itself).  When an unknown
// property is found, it is output to the messages file.  But it is
// also recorded in an array of unknown properties.  This array is
// output to the messages.dat file at the end of the template file.
//
// Note that one parsing functionalility that is not implemented here 
// nor in the grammar is checking the proper syntax of format strings.
// (That is, the suggested format for output of a property value.)
//
// Note that the validation_keycode is not calculated.
//
// There is also an attemnpt to calculate the minimum and maximum
// number of bits in a template.  This adds a fair amount of
// complexity to the code in some places.  For example, keeping
// track of mins and maxs as a consquence of selectcase commands
// requires some careful programming.  Similarly, nested
// structarrays cause some difficulty with bit counting.
//
// Updated October 2003
//
//  1. Corrected bit counting structure.  Did not properly calculate
//     counts for nested selectcases.
//  2. Added verification of data types.  In particular, checks to
//     see if an enumerated value has been defined.
//  3. Added optional comments or eols before first Template statement
//     and optional eols after Validation_Keycode statement.
//  4. Added rudimentary syntax check on the format field.
//
/////////////////////////////////////////////////////////////////////

#include "tdlparser.h"
#include <math.h>

// Token string values.  These are assigned in the grammar.

char* propertyname = NULL;
char* variable_name = NULL;
char* enum_name = NULL;
char* description = NULL;
char* ugid = NULL;
char* format = NULL;
char* physical_unit = NULL;
char* hex_integer = NULL;
char* binary_integer = NULL;
char* abstract = NULL;
char* comment = NULL;
char* assignment_string = NULL;
char* register_mask = NULL;
char* function_propertyname = NULL;  // Used with subproperty function.
char* function_register_mask = NULL; // Used with subproperty function.

// The token numeric values returned by bison (yacc) can not be just 
// by pointer.  They must be copied into a string or else they may be 
// overwritten before we get to the end of the statement. This is not 
// true for the string values above.

char value[MAX_STRING] = "";
char startvalue [MAX_STRING] = "";
char tolerance [MAX_STRING] = "";
char numeric [MAX_STRING] = "";
char boundary [MAX_STRING] = "";
char size [MAX_STRING] = "";
char case_bit_value [MAX_STRING] = "";
char selectcase_number_of_bits [MAX_STRING] = "";
char tdl_version_number [MAX_STRING] = "";
char access_level[MAX_STRING] = "";
char datatype[MAX_STRING] = "";
char subproperty[MAX_STRING] = "";
char validation_keycode[MAX_STRING] = "";
char manufacturer_code [MAX_STRING] = "";
char id_number_of_bits [MAX_STRING] = "";
char template_id [MAX_STRING] = "";

char dot2_1[MAX_STRING] = "";
char dot2_2[MAX_STRING] = "";
char dot2_3[MAX_STRING] = "";
char dot2_4[MAX_STRING] = "";
char dot2_5[MAX_STRING] = "";
char dot2_6[MAX_STRING] = "";
char dot2_7[MAX_STRING] = "";
char dot2_8[MAX_STRING] = "";
char dot2_9[MAX_STRING] = "";
char dot2_10[MAX_STRING] = "";
char dot2_11[MAX_STRING] = "";
char dot2_12[MAX_STRING] = "";

// Non token variables.

bool assignment_found = false;

// Stuff to keep track of template statistics.

int min_number_bits[MAX_TEMPLATES];  
int max_number_bits[MAX_TEMPLATES];
int template_ids[MAX_TEMPLATES];
int manufacturer_codes[MAX_TEMPLATES];
char template_titles[MAX_TEMPLATES][MAX_STRING];
int template_index = 0;

// Keep track of statements that cause uncertainty in bit counts.

bool free_form_found[MAX_TEMPLATES];
bool string_types_found[MAX_TEMPLATES];
bool align_found[MAX_TEMPLATES];

// Stuff to help figure out how many bits for a structarray.

int structarray_level = -1;
int structarray_max_bits[MAX_STRUCTS];
int structarray_size[MAX_STRUCTS];

// Stuff to help figure out how many bits for a selectcase.

int select_level = -1;
int case_number[MAX_SELECTS];
int case_min_bits[MAX_SELECTS][MAX_CASES];
int case_max_bits[MAX_SELECTS][MAX_CASES];
int select_bits[MAX_SELECTS]; // the select_number_of_bits for each level

// If any assignment mismatches its data type, output an error message.

bool assignment_data_type_error = false;

// Flag any invalid format fields.

bool invalid_formats = false;

// Keep track of enumeration types, both known and unknown.
// Note that known_enum_list[n][0] is the enumeration type
// and that known_enum_list[n][k] are the enumeration values.

int enum_type_index = -1; // Used by known_enum_type and check_assignment_data_type
int known_enum_val_index = -1; // Used temparoraly for enumeration lists.

int unknown_enum_index = -1;
char unknown_enum_list[1000][MAX_STRING];
int unknown_enum_val_index = -1;
char unknown_enum_values[1000][MAX_STRING];
int known_enum_index = -1;
char known_enum_list[1000][100][MAX_STRING];

// Keep track of what properties are not known.

int unknown_prop_index = -1;
char unknown_properties[1000][MAX_STRING];

// The property list.

#define MAX_PROPERTIES 2000
char property_list [MAX_PROPERTIES][MAX_STRING];


// function prototypes

void print_bit_count (int, int);


/////////////////////////////////////////////////////////////////////
// initialize_parser 
//
// Initialize the parser.  This is executed prior to
// reading anything from the template file.
//
// global inputs: The property_list.txt file.
//
// global outputs: 
//
//    free_form_found
//    string_types_found
//    align_found
//    property_list
//
/////////////////////////////////////////////////////////////////////
void initialize_parser ()

{

free_form_found[template_index] = false;
string_types_found[template_index] = false;
align_found[template_index] = false;


/////////////////////////////////////////
// Open the property list file
/////////////////////////////////////////

FILE *property_file;
char filename[MAX_STRING];

  strcpy (filename, "property_list.txt");
  if ( (property_file=fopen(filename, "r")) == NULL) {
    printf ("ERROR: TDL Parser Main: Can not open file %s\n", filename);
    fprintf (messages, "ERROR: TDL Parser Main: Can not open file %s\n", filename);
    printf ("Hit any key to exit.");
    getchar();
    exit(1);
  };


/////////////////////////////////////////
// Read in the property list
/////////////////////////////////////////

  int index=0;
  while (fgets (property_list[index], MAX_STRING, property_file) != NULL) {
	  strtok (property_list[index], "\n"); // overwrite carriage return with NULL
    index++;
  }
  strcpy (property_list[index], "END_OF_LIST");

  fprintf (messages, "Property List =\n");
  for (index=0; strcmp(property_list[index], "END_OF_LIST") != NULL; index++) {
	  fprintf (messages, "%d, %s\n", index, property_list[index]);
  }

  fclose (property_file);

} // end initialize_parser

/////////////////////////////////////////////////////////////////////
// finalize_parser 
//
// Finish off anything needed to be done.  This is executed after the
// VALIDATION_KEYCODE.
//
// This outputs template statistics, unknown propeties, unknown
// eunumeration types, and invalid enumeration values.
//
// global inputs: almost everything related to template stats.
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void finalize_parser ()

{
bool syntax_correct = TRUE;

  fprintf (messages, "\nTemplate file parse completed.\n\n");

  ////////////////////////////////////////////////////////////
  // Output template statistics.
  ////////////////////////////////////////////////////////////

  for (int i=0; i<template_index; i++) {

	  fprintf (messages, "ManCode=%d ID=%d MinBits=%d MaxBits=%d ",
             manufacturer_codes[i], template_ids[i], 
             min_number_bits[i], max_number_bits[i]);
	  printf ("ManCode=%d ID=%d MinBits=%d MaxBits=%d\n",
             manufacturer_codes[i], template_ids[i], 
             min_number_bits[i], max_number_bits[i]);

    if (align_found[i]) {
      fprintf (messages, "Align found. ");
    }

    if (free_form_found[i]) {
      fprintf (messages, "Free form statements found. ");
    }

    if (string_types_found[i]) {
      fprintf (messages, "String types found.");
    }

    fprintf (messages, "\n");

  } // end for


  ////////////////////////////////////////////////////////////
  // Output unknown properties.
  ////////////////////////////////////////////////////////////

  if (unknown_prop_index >= 0) {

    syntax_correct = FALSE;
    printf ("\nWARNING: The following properties were not known:\n");
    fprintf (messages, "\nWARNING: The following properties were not known:\n");

    int index = 0;
    while (index <= unknown_prop_index) {

    printf ("%s\n", unknown_properties[index]);
      fprintf (messages, "%s\n", unknown_properties[index]);
      index++;

    } // end while

  } // end if

  ////////////////////////////////////////////////////////////
  // Output unknown eunumeration types.
  ////////////////////////////////////////////////////////////

  if (unknown_enum_index >= 0) {

    syntax_correct = FALSE;
    printf ("\nWARNING: The following enumeration types were not known:\n");
    fprintf (messages, "\nWARNING: The following enumeration types were not known:\n");

    int index = 0;
    while (index <= unknown_enum_index) {

      printf ("%s\n", unknown_enum_list[index]);
      fprintf (messages, "%s\n", unknown_enum_list[index]);
      index++;

    } // end while

  } // end if

  ////////////////////////////////////////////////////////////
  // Output unknown eunumeration values.
  ////////////////////////////////////////////////////////////

  if (unknown_enum_val_index >= 0) {

    syntax_correct = FALSE;
    printf ("\nWARNING: The following enumeration values were not consistent with their type:\n");
    fprintf (messages, "\nWARNING: The following enumeration values were not consistent with their type:\n");

    int index = 0;
    while (index <= unknown_enum_val_index) {

      printf ("%s\n", unknown_enum_values[index]);
      fprintf (messages, "%s\n", unknown_enum_values[index]);
      index++;

    } // end while

  } // end if


  ////////////////////////////////////////////////////////////
  // Indicate assignment data type errors.
  ////////////////////////////////////////////////////////////

  if (assignment_data_type_error) {

    syntax_correct = FALSE;
    printf ("\nWARNING: Some assignments were not consistent with their data type.\n");
    fprintf (messages, "\nWARNING: Some assignments were not consistent with their data type.\n");
    fprintf (messages, "Search for 'NOT CONSISTENT' in this file to find them.\n");

  }

  ////////////////////////////////////////////////////////////
  // Indicate invalid formats.
  ////////////////////////////////////////////////////////////

  if (invalid_formats) {

    syntax_correct = FALSE;
    printf ("\nWARNING: Some formats were invalid.\n");
    fprintf (messages, "\nWARNING: Some formats were invalid.\n");
    fprintf (messages, "Search for 'INVALID FORMAT' in this file to find them.\n");

  }

  if (syntax_correct) {
	  printf ("\nSyntax is correct.\n");
	  fprintf (messages, "\nSyntax is correct.\n");
  }else{
	  printf ("\nSyntax is NOT correct.\n");
	  fprintf (messages, "\nSyntax is NOT correct.\n");
  }

  printf("\nCheck file message.dat for more info.\n\n");

} // end finalize_parser

/////////////////////////////////////////////////////////////////////
// initialize_template 
//
// Set up for a new template.  This is executed prior to
// reading anything from each template.
//
// Just initializes several variables so there's no carry over
// from the last template.
//
// global inputs: none
//
// global outputs: 
//
//		hex_integer
//		binary_integer
//		variable_name
//		assignment_string
//		numeric
//
/////////////////////////////////////////////////////////////////////
void initialize_template ()

{
  hex_integer = NULL;
  binary_integer = NULL;
  variable_name = NULL;
  assignment_string = NULL;
  strcpy (numeric, "");

} // end initialize_template

/////////////////////////////////////////////////////////////////////
// finalize_template 
//
// Finish off anything needed to be done for each template.  
// This is executed after each ENDTEMPLATE.
//
// global inputs: manufacturere_codes, template_ids, min_number_bits,
//                max_number_bits, align_found, free_form_found
//                string_types_found
//
// global outputs:	string_types_found, free_form_found, align_found
//                  messages, messages to message.dat, template_index
//
/////////////////////////////////////////////////////////////////////
void finalize_template ()

{

  fprintf (messages, "Finished reading template with manufacturer code %d and ID = %d\n",
			manufacturer_codes[template_index], template_ids[template_index]);
  fprintf (messages, "Minimum number of bits = %d\n", min_number_bits[template_index]);
  fprintf (messages, "Maximum number of bits = %d\n", max_number_bits[template_index]);

  if (align_found[template_index]) {
    fprintf (messages, "There was an align statement so there is some error in the above number of bits calculated.\n");
  }

  if (free_form_found[template_index]) {
    fprintf (messages, "There were free form statements so there the maximum number of bits is indeterminate.\n");
  }

  if (string_types_found[template_index]) {
    fprintf (messages, "There were string types found which increases the min to max range significantly.\n");
  }

  template_index++;
  free_form_found[template_index] = false;
  string_types_found[template_index] = false;
  align_found[template_index] = false;

} // end finalize_template


/////////////////////////////////////////////////////////////////////
// known_property 
//
// input: property name
//
// global inputs: property_list
//
// global outputs: none
//
// return value: true if the property is on the known property list
//               false otherwise
//
/////////////////////////////////////////////////////////////////////
bool known_property (char property[MAX_STRING])
{
bool known = false;
int index = false;

  while (   stricmp(property_list[index],"END_OF_LIST") != 0
         && !known) {

    if (stricmp(property_list[index], property) == 0) {
      known = true;
    }
    index++;

  } // end while

return (known);

} // end known_property

/////////////////////////////////////////////////////////////////////
// known_enum_type 
//
// input: enumeration type name
//
// global inputs: known_enum_list
//
// global outputs: unknown_enum_list (updated if enum type not known)
//                 unknown_enum_index (updated if enum type not known)
//                 enum_type_index - indicates index of current enum_type
//
// return value: true if the enum type is on the known enum list
//               false otherwise
//
/////////////////////////////////////////////////////////////////////
bool known_enum_type (char *variable_name)
{
bool known = false;

  for ( enum_type_index = 0; 
    enum_type_index <= known_enum_index && !known;
    enum_type_index++) {
    if (stricmp(known_enum_list[enum_type_index][0], variable_name) == 0) {
      known = true;
    }  
  } // end for

  // Want enum_type_index to point to current enum type but it gets
  // increment past it.
  enum_type_index--;

  if (!known) {
    unknown_enum_index++;
    strcpy (unknown_enum_list[unknown_enum_index], variable_name);
  }
  
return (known);

} // end known_enum_type

/////////////////////////////////////////////////////////////////////
// valid_bibin
//
// Checks to see if a string in an assigned property is a valid bitbin.
// A valid bitbin can only have 0s, 1s, xs, and commas in it.
//
// inputs: bitbin
//
// outputs: none
//
// return value: true if a valid bitbin
//               false otherwise.
//
/////////////////////////////////////////////////////////////////////
bool valid_bitbin (char* bitbin)
{
bool valid = true;


  int lastchar = strlen(bitbin)-1;
  if (   bitbin[0] == '\"' 
      && bitbin[lastchar] == '\"') {

    for (int i=1; i<lastchar; i++) {
      if (   bitbin[i] != '0'
          && bitbin[i] != '1'
          && bitbin[i] != 'x'
          && bitbin[i] != 'X'
          && bitbin[i] != ',' ) {

        valid = false;

      } // end if
    } // end for

  } // end if quoted

return (valid);

} // end valid_bitbin

/////////////////////////////////////////////////////////////////////
// valid_format
//
// Does a rudimentary check to see if a string in a property format 
// field is valid.  All it does is check to see that there are no
// illegal characters.  It does not verify that the characters
// are in a meaningful sequence.
// 
// inputs: format
//
// outputs: none
//
// return value: true if a valid format
//               false otherwise.
//
/////////////////////////////////////////////////////////////////////
bool valid_format (char* format)
{
bool valid = true;

  int lastchar = strlen(format)-1;
  if (   format[0] == '\"' 
      && format[lastchar] == '\"') {

    for (int i=1; i<lastchar; i++) {
      if (   format[i] != '$'
          && format[i] != '%'
          && format[i] != 'E'
          && format[i] != 'e'
          && format[i] != '+'
          && format[i] != '-'
          && format[i] != '.'
          && format[i] != ','
          && format[i] != '#'
          && format[i] != '0'
          && format[i] != 'R'
          && format[i] != 'r'
          && format[i] != 'P'
          && format[i] != 'p'
          && format[i] != 'S'
          && format[i] != 's'
          && format[i] != '/'
          && format[i] != 'D'
          && format[i] != 'd'
          && format[i] != 'M'
          && format[i] != 'm'
          && format[i] != 'Y'
          && format[i] != 'y' ) {

        valid = false;

      } // end if
    } // end for

  } // end if quoted

return (valid);

} // end valid_format

/////////////////////////////////////////////////////////////////////
// valid_enum_cardinal
//
// Checks to see if a numeric assignment for an enumeration type
// is within bounds for the enumeration.
//
// inputs: cardinal
//
// outputs: none
//
// return value: true if a valid cardinal for the enumeration type
//               false otherwise.
//
/////////////////////////////////////////////////////////////////////
bool valid_enum_cardinal (char cardinal[MAX_STRING])
{
bool valid = true;

  int int_cardinal = atoi(cardinal);

  if (int_cardinal != atof(cardinal)) {
    
    // Not an integer.
    valid = false;

  }else{

    // Find max valid cardinal
    int max = 1;
    while (stricmp(known_enum_list[enum_type_index][max], "END_OF_LIST") != 0) {
      max++;
    } // end while

    if (int_cardinal < 0 || max <= int_cardinal) {
      valid = false;
    } // end if
  
  } // end if-then-else

return (valid);

} // end valid_enum_cardinal

/////////////////////////////////////////////////////////////////////
// check_assignment_data_type 
//
// An assignment has been made to a property command. Check to see 
// if the assignment is consistent with the declared data type.
//
// global inputs: hex_integer, binary_integer, variable_name,
//                assignment_string, numeric, datatype, 
//                unknown_enum_values
//
// global outputs: assignment_data_type_error
//
/////////////////////////////////////////////////////////////////////
void check_assignment_data_type ()
{
char *value;
bool known = false;

  // Every assignment is stored as string, so transfer to common variable.

  if (hex_integer != NULL) {
    value = hex_integer;
  }else if (binary_integer != NULL) {
    value = binary_integer;
  }else if (variable_name != NULL) {
    value = variable_name;
  }else if (assignment_string != NULL) {
    value = assignment_string;
  }else{
    value = numeric;
  } // end if hex_integer

  // Check each datatype one at a time.

  if (   hex_integer != NULL
      || binary_integer != NULL) {

	  // Any assignment can be hex or binary since this imitates
	  // what would be read from the TEDS.

	  known = true;

  }else if (   (	 (stricmp (datatype, "CHR5") == 0)
                || (stricmp (datatype, "ASCII") == 0) 
                || (stricmp (datatype, "Unicode") == 0) 
                || (stricmp (datatype, "Sting5") == 0) 
                || (stricmp (datatype, "String7") == 0) 
                || (stricmp (datatype, "String16") == 0) )
            && (assignment_string != NULL) ) {

    known = true;

  }else if (   (	 (stricmp (datatype, "ConRelRes") == 0)
                || (stricmp (datatype, "ConRes") == 0)
                || (stricmp (datatype, "Single") == 0) )
            && (stricmp (numeric, "") != 0) ) {

    known = true;

  }else if (   (stricmp (datatype, "Date") == 0)
            && (stricmp (numeric, "") != 0) 
            && (atoi(value) == atof(value)) ) { // is an int

    known = true;

  }else if (   (stricmp (datatype, "UNINT") == 0)
            && (stricmp (numeric, "") != 0) 
            && (atoi(value) == atof(value)) ) { // is an int

    known = true;

  }else if (   (stricmp (datatype, "BitBin") == 0)
            && (assignment_string != NULL)
            && (valid_bitbin(value)) ) {

    known = true;

  }else if (   (stricmp (datatype, "Enumeration") == 0)
            && (stricmp (numeric, "") != 0) 
            && (valid_enum_cardinal(numeric)) ){

    // It's an enumeration that has been assigned based on its cardinal.

    known = true;

  }else if (   (stricmp (datatype, "Enumeration") == 0)
            && (enum_type_index <= known_enum_index) ) {

    // It's an enumeration and we know the enum type.
    // Check to see if the value provided is one of the valid types
    // for this enumeration.

    int index = 1;
    while (   stricmp(known_enum_list[enum_type_index][index], "END_OF_LIST") != 0
           && !known) {

     if (stricmp(known_enum_list[enum_type_index][index], value) == 0) {
        known = true;
      }
      index++;

    } // end while

    if (!known) {
      unknown_enum_val_index++;
      strcpy(unknown_enum_values[unknown_enum_val_index], value);
    }

  } // end if stricmp(datatype ...

  if (!known) {
    fprintf (messages, "WARNING: This is NOT CONSISTENT with the stated data type.\n");
    assignment_data_type_error = true;
  }

} // end check_assignment_data_type 

/////////////////////////////////////////////////////////////////////
// process_property_definition 
//
// A property definition command reads a value from
// the TEDS and assigns it to the specified property.
// Alternatively, the value can be assigned directly.
//
// Syntax:
//
// %<property_tag>,<"description">,<access_level>,<data_type>,<format>,<physical_unit>
// or
// %<property_tag>,<"description">,<access_level>,<data_type>,<format>,<physical_unit> = <value>
// or
// %<property_tag><[subproperty]>,<"description">,<access_level>,<data_type>,<format>,
// <physical_unit>
// or
// %<property_tag><[subproperty]>,<"description">,<access_level>,<data_type>,<format>,
// <physical_unit> = <value>
// 
// For the purposes of the assignment (whether read from the TEDS or
// given explicitely), the description, access_level, and physical_unit are
// extraneous.  The size tells the number of bits to read.  The datatype
// indicates what type of conversion of the bits is necessary.
//
// This routine does quite a lot.  However, it is broken down into
// independent sequential activities as indicated by the major comment headers.
// Mostly it just outputs messages indicating the values of the properties.
// However, it also does some of the bit counting stuff.
//
// global Inputs:	
//
//    propertyname, description, access_level, size, datatype,
//    physical_unit, hex_integer, binary_integer, messages, format
//    dot2_1 ... dot2_12, variable_name, assignment_string, numeric,
//    structarray_bit_count, unknown_prop_index, unknown_properties
//
// global outputs: 
//
//    string_types_found, unknown_prop_index, hex_integer,
//    binary_integer, variable_name, numeric, assignment_string,
//    dot2_1, comments to the message file
//
/////////////////////////////////////////////////////////////////////
void process_property_definition ()

{
int local_bit_count = 0;
	
  ///////////////////////////////////////////////////////
  // General for all properties
  ///////////////////////////////////////////////////////

  fprintf (messages, "The property %s was just defined.\n", propertyname);

  if (description != NULL) {
    fprintf (messages, "It was described as:\n %s\n", description);
  }else{
    fprintf (messages, "It has a function subproperty association with: %s[%s].\n",
   	function_propertyname, function_register_mask);
  }

  fprintf (messages, "It has access level %s", access_level);

  if (format != NULL) {
    fprintf (messages, " and format %s,\n", format);
    if (!valid_format(format)) {
      fprintf (messages,"WARNING: INVALID FORMAT.\n");
      invalid_formats = true;
    }
  }else{
    fprintf (messages, " and no format specified,\n");
  }

  if (physical_unit != NULL) {
    fprintf (messages, " and units %s.\n", physical_unit);
  }else if (stricmp(dot2_1, "") != 0) {
    fprintf (messages, " and units (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s).\n",
             dot2_1, dot2_2, dot2_3, dot2_4, dot2_5, dot2_6, dot2_7, dot2_8, 
             dot2_9, dot2_10, dot2_11, dot2_12);
  }else{
    fprintf (messages, " and no units specified.\n");
  }


  ///////////////////////////////////////////////////////
  // Different datatypes need to be handled differently.
  ///////////////////////////////////////////////////////

  if (	 (stricmp (datatype, "ConRelRes") == 0)
      || (stricmp (datatype, "ConRes") == 0) ) {

    fprintf (messages, "The data type is %s with start value %s and tolerance %s.\n", datatype, startvalue, tolerance);

  }else if (stricmp (datatype, "Enumeration") == 0) {

    fprintf (messages, "The data type is an enumeration of %s.\n", enum_name);
    if (!known_enum_type(enum_name)) {
      fprintf (messages, "WARNING: This is an UNKNOWN Enumeration Type!\n");
    }

  }else{

    fprintf (messages, "The data type is %s.\n", datatype);

  }

  ///////////////////////////////////////////////////////
  // Is it an assignment or do you read from the TEDS?
  // Also increment min and max bits read variable.
  ///////////////////////////////////////////////////////

  int integer_size = atoi(size);
  fprintf (messages, "The size parameter is %d\n", integer_size);

  if (!assignment_found) {

    // It's not an assignment. Read the bits from the TEDS.

    fprintf (messages, "The value of %s should be read from the TEDS.\n", propertyname, size);

    ///////////////////////////////////////////////////////
    // If it's a string type, have to multply the size by char length to get max
    // bits read. The min bits read is 0.
    ///////////////////////////////////////////////////////
	 
    if (stricmp (datatype, "string5") == 0) {

      string_types_found[template_index] = true;
      local_bit_count = (pow(2, integer_size) - 1.0)*5;

    }else if (stricmp (datatype, "string7") == 0) {

      string_types_found[template_index] = true;
      local_bit_count = (pow(2, integer_size) - 1.0)*7;

    }else if (stricmp (datatype, "string16") == 0) {

      string_types_found[template_index] = true;
      local_bit_count = (pow(2, integer_size) - 1.0)*16;

    }else{ // not a string type

      local_bit_count = integer_size;

    } // end if-then-else string type


    if (structarray_level > -1) { 
      // we're in a structarray
      structarray_max_bits[structarray_level] += local_bit_count;
    }else{
      // we're not in a structarray
      print_bit_count (local_bit_count, local_bit_count);
    } // end if structarray_level


  }else{ // It's an assignment.

    if (hex_integer != NULL) {
      fprintf (messages, "The value of %s is stated as the hex value %s.\n", propertyname, hex_integer);
    }else if (binary_integer != NULL) {
      fprintf (messages, "The value of %s is stated as the binary value %s.\n", propertyname, binary_integer);
    }else if (variable_name != NULL) {
      fprintf (messages, "The value of %s is stated as the string value %s.\n", propertyname, variable_name);
    }else if (assignment_string != NULL) {
      fprintf (messages, "The value of %s is stated as the string value %s.\n", propertyname, assignment_string);
    }else{
      fprintf (messages, "The value of %s is stated as the decimal value %s.\n", propertyname, numeric);
    } // end if hex_integer

    check_assignment_data_type();

  } // end if !assignment

  ///////////////////////////////////////////////////////
  // Free form properties are special.
  ///////////////////////////////////////////////////////

  if (   (   stricmp(propertyname,"%user")      == 0 
          || stricmp(propertyname,"%asciiteds") == 0
          || stricmp(propertyname,"%xml")       == 0)
       && stricmp(size,"0") == 0) {

    fprintf (messages, "Since the bit size for this free form command is 0, the rest of the TEDS is user defined.\n");
    fprintf (messages, "The bits read from the TEDS should be treated as %s.\n", datatype);
    free_form_found[template_index] = true;

  }else if (!known_property(propertyname)) {

    fprintf (messages, "WARNING: Property %s is not on the known property list!\n", propertyname);
    unknown_prop_index++;
    strcpy (unknown_properties[unknown_prop_index], propertyname);

  } // end if %user

  ///////////////////////////////////////////////////////
  // cleanup
  ///////////////////////////////////////////////////////

  hex_integer = NULL;
  binary_integer = NULL;
  variable_name = NULL;
  strcpy (numeric, "");
  assignment_string = NULL;
  strcpy (dot2_1, "");

} // end process_property_definition 

/////////////////////////////////////////////////////////////////////
// process_template_statement 
//
// Syntax:
//
//	TEMPLATE <manufacturer code>, <ID number of bits>, <Template ID>,<title>
//		<commands>
//	ENDTEMPLATE
//
// global inputs: messages, template_string, manufacturer_code
//                template_id, description, id_number_of_bits,
//                template_index
//
// global outputs: messages to message.dat, manufacturer_code,
//                 template_ids, template_titles
//
/////////////////////////////////////////////////////////////////////
void process_template_statement ()

{

  fprintf (messages, "\nJust read Template statement.\n%");
  fprintf (messages, "Manufacturer code = %s\n", manufacturer_code);
  fprintf (messages, "Template ID = %s\n", template_id);
  fprintf (messages, "Title = %s\n", description);
  print_bit_count (atoi(id_number_of_bits), atoi(id_number_of_bits));

  manufacturer_codes[template_index] = atoi(manufacturer_code);
  template_ids[template_index] = atoi(template_id);
  strcpy(template_titles[template_index], description);

} // end process_template_statement

/////////////////////////////////////////////////////////////////////
// process_abstract_statement
//
// Syntax:
//
//	ABSTRACT <description>
//
// global inputs: messages, abstract
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_abstract_statement ()

{

  fprintf (messages, "Just read abstract (no bits read):\n%s\n", abstract);

} // end process_abstract_statement

/////////////////////////////////////////////////////////////////////
// process_align
//
// Syntax:
//
//	ALIGN <word_size>
//
// The tricky part about this is calculating the bits to skip.
// Since this is dependent on how many bits have already been read,
// this can not be determined without actually reading a TEDS.
// Thus, the bit counts have an error for any template containing
// an align command.
//
// global inputs: messages, boundary
//
// global outputs: messages to message.dat, align_found
//
/////////////////////////////////////////////////////////////////////
void process_align ()

{

int memory_size = 255;
int boundary_int = atoi(boundary);
// current position, X, is the next bit to be read so need to add 1 to total_bits_read)
int bits_to_skip = (memory_size - (min_number_bits[template_index]+1) + 1) % boundary_int;

  fprintf (messages, "Time to align the next bits to be read as if we had %d sized bytes.\n", boundary_int);
  fprintf (messages, "As an illustration, assume the total memory size, Y, is %d\n", memory_size);
  fprintf (messages, "We've already read %d bits (call this X).\n", min_number_bits);
  fprintf (messages, "So calculate (Y-X+1) MOD# = (%d-%d+1)mod%d and skip %d bits.\n", 
           memory_size, min_number_bits[template_index]+1, boundary_int, bits_to_skip);

  print_bit_count (bits_to_skip, bits_to_skip);
  align_found[template_index] = true;

} // end process_align

/////////////////////////////////////////////////////////////////////
// process_comment
//
// Syntax:
//
//	// <description>
//
// global inputs: messages, comment
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_comment ()

{
  fprintf (messages, "Just read comment (no bits read):\n%s\n", comment);

} // end process_comment

/////////////////////////////////////////////////////////////////////
// process_enumerate
//
// Syntax:
//
//	Enumerate <defined_type>, <enum_list>
//
// global inputs: messages, enum_name
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_enumerate ()

{
  fprintf (messages, "The Enumerate class %s is defined as:\n", variable_name);
  known_enum_index++;
  strcpy(known_enum_list[known_enum_index][0], enum_name);
  known_enum_val_index = 0;

} // end process_enumerate

/////////////////////////////////////////////////////////////////////
// process_enumerate_item
//
// global inputs: messages, numeric, variable_name
//
// global outputs: messages to message.dat, knonw_enum_list
//
/////////////////////////////////////////////////////////////////////
void process_enumerate_item ()

{
  if (variable_name != NULL) {
    fprintf (messages, "%s\n", variable_name);
    known_enum_val_index++;
    strcpy(known_enum_list[known_enum_index][known_enum_val_index], 
           variable_name);
  }else{
    fprintf (messages, "%s\n", numeric);
  }

} // end process_enumerate_item

/////////////////////////////////////////////////////////////////////
// process_enumerate
//
// global inputs: messages, enum_name
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_end_enumerate ()

{
  // Need to mark end of each list.

  known_enum_val_index++;
  strcpy(known_enum_list[known_enum_index][known_enum_val_index], 
         "END_OF_LIST");

} // end process_end_enumerate

/////////////////////////////////////////////////////////////////////
// process_physical_unit
//
// global inputs: messages, variable_name, dot2_1 ... dot2_12
//
// global outputs: messages to message.dat, dot2_1
//
/////////////////////////////////////////////////////////////////////
void process_physical_unit ()

{
  fprintf (messages, "Physical_unit %s has been associated with", 
           variable_name);
  fprintf (messages, " (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s).\n",
           dot2_1, dot2_2, dot2_3, dot2_4, dot2_5, dot2_6, dot2_7, dot2_8, 
           dot2_9, dot2_10, dot2_11, dot2_12);

  strcpy (dot2_1, "");

} // end process_physical_unit


/////////////////////////////////////////////////////////////////////
// process_selectcase
//
// Syntax:
//
//	SELECTCASE <"description">, <access_level>, <number_of_bits> 
//		CASE <"description">, <bit value> 
//			<command list>
//		ENDCASE
//		...
//		CASE <"description">, <bit value> 
//			<command list>
//		ENDCASE
//	ENDSELECT 

// global inputs: messages, description, access_level,
//                selectcase_number_of_bits, select_level
//
// global outputs: messages to message.dat, select_bits, case_number,
//                 select_level
//
/////////////////////////////////////////////////////////////////////
void process_selectcase ()

{

  select_level++;

  fprintf (messages, "A selectcase statement with description %s\n and access level %s has been read.\n",
           description, access_level);
  fprintf (messages, "The next %s bits from the TEDS are the selectcase value for the following cases.\n",
           selectcase_number_of_bits);
  select_bits[select_level] = atoi(selectcase_number_of_bits);
  case_number[select_level] = -1;

} // end process_selectcase

/////////////////////////////////////////////////////////////////////
// process_endselect
//
// The complexity here is that you have to figure out min and max
// bits based on all possible combinations of cases and nested
// selectcase statements.  The basic approach is to keep track
// of mins and maxs for each case and for each selectcase level.
// This is essentially a recursive routine implemented with arrays.
//
// global inputs: messages, case_min_bits, case_nax_bits, select_level
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_endselect ()

{

  fprintf (messages, "End of selectcase statement.\n");

  // Find minimum bits out of all cases.

  int min_bits = 100000;
  for (int i=0; i<=case_number[select_level]; i++) {
    if (min_bits > case_min_bits[select_level][i]) {
      min_bits = case_min_bits[select_level][i];
    }
  } // end for

  // Find maximum bits out of all cases.

  int max_bits = 0;
  for (i=0; i<=case_number[select_level]; i++) {
    if (max_bits < case_max_bits[select_level][i]) {
      max_bits = case_max_bits[select_level][i];
    }
  } // end for

  // Select_level is decremented because print_bit_count uses it to
  // decide whether it's the top level select or not.
  // But we still have to add in the number of bits read in the selectcase.
  // So the select_bits index has to be upped by 1.
  
  select_level--;
  print_bit_count(min_bits+select_bits[select_level+1], max_bits+select_bits[select_level+1]);

} // end process_endselect

/////////////////////////////////////////////////////////////////////
// process_case
//
// global inputs: messages, description, case_bit_value, case_number,
//                select_level
//
// global outputs: messages to message.dat, case_number, case_min_bits,
//                 case_max_bits
//
/////////////////////////////////////////////////////////////////////
void process_case ()

{

  fprintf (messages, "Case %s or bit value %s. No bits read.\n", description, case_bit_value);

  case_number[select_level]++;
  case_min_bits[select_level][case_number[select_level]] = 0;
  case_max_bits[select_level][case_number[select_level]] = 0;

} // end process_case

/////////////////////////////////////////////////////////////////////
// process_case_end
//
// global inputs: messages
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_case_end ()

{

  fprintf (messages, "End of case statement.\n");
  fprintf (messages, "This case reads between %d and %d bits.\n",
           case_min_bits[select_level][case_number[select_level]],
           case_max_bits[select_level][case_number[select_level]]);

} // end process_case_end

/////////////////////////////////////////////////////////////////////
// process_spacing
//
// Syntax:
//
//	spacing
//
// global inputs: messages
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_spacing ()

{

  fprintf (messages, "Spacing statement (no bits read).\n");

} // end process_spacing

/////////////////////////////////////////////////////////////////////
// process_structarray_statement
//
// This is slightly complicated by having to track number of bits.
//
// Syntax:
//
//	STRUCTARRAY <name>, <"description">, <access_level>, <number_of_bits> 
//		<property_list>
//	ENDSTRUCTARRAY
//
// global inputs: messages, variable_name, description, access_level
//                size, structarray_level
//
// global outputs: messages to message.dat, structarray_level
//                 struct_array_max_bits, structarray_size
//
/////////////////////////////////////////////////////////////////////
void process_structarray_statement ()

{

  fprintf (messages, "Structarray %s. with description \n%s\n", variable_name, description);
  fprintf (messages, "and Access Level %s\n", access_level);
  
  structarray_level++;
  structarray_max_bits[structarray_level] = 0;
  structarray_size[structarray_level] = atoi(size);

} // end process_structarray_statemnt

/////////////////////////////////////////////////////////////////////
// structarray_bits 
//
// Calculates the maximum number of bits a structarray can use.
// Since the structarray command indicates the number of bits to be
// read to determine how many elements are in the array, the max bits
// is dependent on raising the 2 to the number of bits.
//
// global inputs: structarray_size, structarray_max_bits
//
// global outputs: none
//
// Returns max number of bits for the structarray.
//
/////////////////////////////////////////////////////////////////////
int structarray_bits (int level)
{

return ((pow(2, structarray_size[level]) - 1.0)*structarray_max_bits[level]
         + structarray_size[level]);

} // end structarray_bits

/////////////////////////////////////////////////////////////////////
// process_endstructarray_statement 
//
// Complicated by calculating number of bits and nested structarrays.
//
// global inputs: messages, structarray_size, structarray_bit_count,
//                structarray_level
//
// global outputs: messages to message.dat, structarray_max_bits,
//                 structarray_level
//
/////////////////////////////////////////////////////////////////////
void process_endstructarray_statement ()

{

  fprintf (messages, "EndStructarray\n");


  // Have to increment max bits by number of possible array elements. 

  if (structarray_level != 0) {

    // We're in a nested structarray.
    structarray_max_bits[structarray_level-1] += structarray_bits(structarray_level);

  }else if (structarray_level == 0) {

    // This is top level structarray.
    print_bit_count (structarray_size[0], structarray_bits(0));

  }

  structarray_level--;

} // end process_endstructarray_statement

/////////////////////////////////////////////////////////////////////
// process_subproperty
//
// This is part of the property command.
//
// global inputs: messages, binary_integer, hex_integer, subproperty
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_subproperty ()

{

  if (register_mask == 0) {
    fprintf (messages, "Subproperty %s\n", subproperty);
  }else{
    fprintf (messages, "Subproperty register mask %s\n", register_mask);
  }

} // end process_subproperty

/////////////////////////////////////////////////////////////////////
//
//	TDL_VERSION_NUMBER <version number>
//
// Syntax:
//
//	TDL_VERSION_NUMBER <version number>
//
// global inputs: messages,	tdl_version_number
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_tdl_version_number ()

{

  fprintf (messages, "The TDL Version Number is %s.  No bits read.\n", tdl_version_number);

} // end tdl version number

/////////////////////////////////////////////////////////////////////
// process_ugid
//
// Syntax:
//
//	UGID <identifier>, <"description">
//
// global inputs: messages, ugid
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_ugid ()

{

  fprintf (messages, "The UGID %s is described by %s.  No bits read.\n", 
           ugid, description);

} // end process_ugid

/////////////////////////////////////////////////////////////////////
// process_validation_keycode
//
// Syntax:
//
//	VALIDATION_KEYCODE <checksum>
//
// global inputs: messages, validation_keycode
//
// global outputs: messages to message.dat
//
/////////////////////////////////////////////////////////////////////
void process_validation_keycode ()

{

  fprintf (messages, "The validation keycode is %s.  No bits read.\n", validation_keycode);

} // end validation_keycode

/////////////////////////////////////////////////////////////////////
// print_bit_count (a debug routine)
//
// This not only prints the bit count but also does some of the
// logic to count min and max bits.  In particular, it helps with
// nested selectcases.  It also updates the template bit count array
// so they can be output at the end of the parse.
//
// global inputs: messages, min_bits_read, max_bits_read,
//                case_min_bits, case_max_bits, template_index,
//                min_number_bits, max_number_bits
//
// global outputs: messages to message.dat, min_number_bits, 
//                 max_number_bits, case_min_bits, case_max_bits,
//                 min_number_bits, max_number_bits
//
/////////////////////////////////////////////////////////////////////
void print_bit_count (int min_bits_read, int max_bits_read)
{

  if (select_level < 0) {

    // Not in a selectcase so update bits read.

    min_number_bits[template_index] += min_bits_read;
    max_number_bits[template_index] += max_bits_read;

    fprintf (messages, "Just read between %d and %d bits.\n", 
             min_bits_read, max_bits_read);
    fprintf (messages, "Total bits read between %d and %d\n", 
             min_number_bits[template_index], max_number_bits[template_index]);

  }else{

    // In a selectcase so update case bit counts.

    fprintf (messages, "Potentially read between %d and %d bits.\n", 
             min_bits_read, max_bits_read);

    case_min_bits[select_level][case_number[select_level]] += min_bits_read;
    case_max_bits[select_level][case_number[select_level]] += max_bits_read;

  }

} // end print_bit_count

/////////////////////////////////////////////////////////////////////
// print_token (a debug routine)
//
// global inputs: tokens, token
//
// global outputs: token printed to tokens.dat
//
/////////////////////////////////////////////////////////////////////
void print_token (char* token)
{

  fprintf (tokens, "%s\n", token);
  fflush (tokens);

} // end print_token

/////////////////////////////////////////////////////////////////////
// print_statement_num (a debug routine)
//
// global inputs: messages, tokens, statement_count
//
// global outputs: messages to tokens.dat and messages.dat
//
/////////////////////////////////////////////////////////////////////
void print_statement_num ()
{
  printf ("Line: %d\n", statement_count);
  fprintf (tokens, "Line : %d\n", statement_count);
  fprintf (messages, "\nLine: %d\n", statement_count);

} // end print_statement_num
